/*****************************************************************************
 *
 * File:		test_fill.cpp
 *
 * Author:		Sven Mulka
 *
 * Company:		Ingenieurbro Mulka, Software & Data Communication
 *				Copyright (c) 2018, by IB-Mulka
 *				http://www.ib-mulka.de
 *
 * Content:		Test Application to trigger $fill issue of linker.
 *
 *				The $fill issue will be triggered by ConvertIntFloat().
 *				Must be compiled with --speculate_loads=auto
 *
 *				The compiler will add build attributes
 *
 *					.battr "TI", Tag_File, 1, Tag_pre_pad_bytes(48)
 *					.battr "TI", Tag_File, 1, Tag_post_pad_bytes(48)
 *
 *				Look in intermediate file test_fill.asm for more details.
 *
 * Changes:
 * 17.07.18 SM  Initial Version
 * 07.08.18 SM	Add function ConvertIntFloat() to trigger $fill issue.
 *
 *****************************************************************************/
/*
 * $Log:   $
 */
#include <c6x.h>                        // C66xx intrinsics
#include <c6x_vec.h>                    // OpenCL support for C66xx

#include <iostream>
#include <iomanip>
#include <complex>

using namespace std;

/*****************************************************************************/

typedef unsigned  __short2_t;

typedef float				float_t;	// the float type
typedef complex<float_t>	cfloat_t;	// a complex float type
typedef complex<int16_t>	cint16_t;	// a complex int16 type

template <typename T, size_t N>
inline size_t countof(T const (&arr)[N]) { return N; }

/*****************************************************************************/

//#pragma DATA_SECTION (".omp_data")
cint16_t g_data_i[256] = { cint16_t(+256, 0) };

//#pragma DATA_SECTION (".omp_data")
cfloat_t g_data_a[256] = { cfloat_t(+1.0f, 0.0f) };

//#pragma DATA_SECTION (".omp_data")
cfloat_t g_data_b[256] = { cfloat_t(+2.0f, 1.0f) };

/*****************************************************************************
 * Copy a complex float array to another.
 */
void CopyArray(
	cfloat_t 		out[],
	const cfloat_t in[],
	size_t			count
	)
{
#ifdef __USE_TMS320C6600__
	const __float2_t* restrict pin = (const __float2_t*)in;
	__float2_t* restrict pout = (__float2_t*)out;
	int idx = 0;
	int end = (int)(count / 6) * 6;

	/***
	 * Runs 6 copy operation in the loop.
	 *
	 * ii = 6  Schedule found with 3 iterations in parallel (no SPLOOP)
	 * Loop Unroll Multiple : 6x
	 * Total cycles (est.) : 7 + trip_cnt * 6
	 *
	 * Minimum required memory pad   : 48 bytes
	 * Minimum threshold value       : -mh96
	 */
	DATA_IS_ALIGNED_8(pin);
	DATA_IS_ALIGNED_8(pout);
	#pragma MUST_ITERATE (0,,6)
	for (; idx < end; idx++)
	{
		_amem8_f2(pout++) = _amem8_f2_const(pin++);
	}

	/***
	 * Runs the copy operation in the loop.
	 *
	 * ii = 1  Schedule found with 7 iterations in parallel (no SPLOOP)
	 * Total cycles (est.) : 6 + trip_cnt * 1 = Between 7 and 11
	 */
	DATA_IS_ALIGNED_8(pin);
	DATA_IS_ALIGNED_8(pout);
	#pragma MUST_ITERATE (0, 5)
	for (; idx < (int)count; idx++)
	{
		_amem8_f2(pout++) = _amem8_f2_const(pin++);
	}
#else
	for (int idx = 0; idx < (int)count; idx++)
	{
		out[idx] = in[idx];
	}
#endif

}

/*****************************************************************************
 * Converts IQ samples from complex int16 to complex float.
 * Works also inplace and runs from the back to the front.
 *
 * Normally the scale factor should be (1.0f / 32767.0f) to convert a range
 * of (-32767 .. +32767) to (-1.0 .. +1.0)
 */
void ConvertIntFloat(
	cfloat_t		out[],
	const cint16_t	in[],
	size_t			count,
	float			scale
	)
{
#ifdef __USE_TMS320C6600__
	const __short2_t* restrict ps = ((const __short2_t*)in) + count - 1;
	__float2_t* restrict pd = ((__float2_t*)out) + count - 1;
	__float2_t  sc = _ftof2(scale, scale);

	int idx = 0;
	int end = ((int)count/6)*6;

	/***
	 * Runs 6 conversion int16 to float in the loop.
	 *
	 * ii = 7  Schedule found with 3 iterations in parallel (no SPLOOP)
	 * Loop Unroll Multiple : 6x
	 * Total cycles (est.) : 12 + trip_cnt * 7
	 */
	DATA_IS_ALIGNED_4(ps);
	DATA_IS_ALIGNED_8(pd);
	#pragma MUST_ITERATE (0,,6)
	for (; idx < end; idx++)
	{
		_amem8_f2(pd--) = _dmpysp(sc, _dinthsp(_amem4_const(ps--)));
	}

	/***
	 * Runs the conversion int16 to float in the loop.
	 *
	 * ii = 1  Schedule found with 13 iterations in parallel (no SPLOOP)
	 * Total cycles (est.) : 12 + trip_cnt * 1 = Between 13 and 17
	 */
	DATA_IS_ALIGNED_4(ps);
	DATA_IS_ALIGNED_8(pd);
	#pragma MUST_ITERATE (0,5)
	for (; idx < count; idx++)
	{
		_amem8_f2(pd--) = _dmpysp(sc, _dinthsp(_amem4_const(ps--)));
	}
#else
	const cint16_t*  ps = in + count - 1;
	cfloat_t* pd = out + count - 1;
	int idx = 0;
	for (int idx = 0; idx < count; idx++)
	{
		cint16_t v = *ps--;
		*pd-- = scale * cfloat_t(v.real(), v.imag());
	}
#endif
}

/*****************************************************************************
 * Run our own optimized function to trigger $fill issue.
 */
bool test_fill()
{
	using namespace std;

	cout << endl
		 << "********************************************************" << endl
		 << "Running test.fill ..." << endl << endl;

	ConvertIntFloat(g_data_a, g_data_i, countof(g_data_i), 1.0f/256);

	cout << "re[0] before : " << g_data_b[0].real() << endl;

	CopyArray(g_data_b, g_data_a, countof(g_data_a));

	cout << "re[0] after : " << g_data_b[0].real() << endl;

	cout << endl
		 << "********************************************************" << endl
		 << endl
		 ;
	return true;
}

/*****************************************************************************/
